/* vi:ts=4:sw=4
 *
 * VIM - Vi IMitation
 *
 * Code Contributions By:	Bram Moolenaar			mool@oce.nl
 *							Tim Thompson			twitch!tjt
 *							Tony Andrews			onecom!wldrdg!tony 
 *							G. R. (Fred) Walter		watmath!watcgl!grwalter 
 */

/*
 * screen.c: code for displaying on the screen
 */

#include "vim.h"
#include "globals.h"
#include "proto.h"
#include "param.h"
#ifdef KANJI
#include "kanji.h"
#endif

/*
	Macros handling inverted lines
*/
#define	ScreenAttr(screenp)	((screenp) - Nextscreen + Nextattrib)
#define	ScreenSetAttr(x)	(*x = 0x0)
#define	ScreenClrAttr(x)	(*x = 0x0)
#define		AM_None			0x00
#define		AM_Inv			0x02
#ifdef FEP
# define	AM_UL			0x04
#endif /* FEP */
#ifdef KANJI
# define	AM_K1			0x01
#endif /* KANJI */

char *tgoto __PARMS((char *cm, int col, int line));

static u_char 	*Nextscreen = NULL; 	/* What's to be put on the screen. */
static u_char	*Nextattrib = NULL; 	/* Attributes for Nextscreen.      */
static int		 NumLineSizes = 0;		/* # of active LineSizes */
static linenr_t *LineNumbers = NULL;	/* Pointer to the line for LineSizes */
static u_char 	*LineSizes = NULL;		/* Number of rows the lines occupy */
static u_char 	**LinePointers = NULL;	/* array of pointers into Netscreen */

#ifdef FEP
static int uline_off(), uline_on(); 
#endif
#ifdef KANJI
static int kanji_off(), kanji_on(); 
#endif
static int invert_off(), invert_on(); 

/*
 * The following variable is set (in cursupdate) to the number of physical
 * lines taken by the line the cursor is on. We use this to avoid extra calls
 * to plines(). The optimized routine updateline()
 * makes sure that the size of the cursor line hasn't changed. If so, lines
 * below the cursor will move up or down and we need to call the routine
 * updateScreen() to examine the entire screen.
 */
static int		Cline_size; 			/* size (in rows) of the cursor line */
static int		Cline_row;				/* starting row of the cursor line */
int				redraw_msg = TRUE;		/* TRUE when "insert mode" needs updating */
static FPOS		oldCurpos = {0, 0};		/* last known end of quoted part */
static int		oldCurswant = 0;		/* last known value of Curswant */
static int		canopt;					/* TRUE when cursor goto can be optimized */
static int		emptyrows = 0;			/* number of '~' rows on screen */

static int screenline __ARGS((linenr_t, int, int));
static void screenchar __ARGS((u_char *, int, int));
static void screenfill __ARGS((int, int));

/*
 * updateline() - like updateScreen() but only for cursor line
 *
 * This determines whether or not we need to call updateScreen() to examine
 * the entire screen for changes. This occurs if the size of the cursor line
 * (in rows) hasn't changed.
 */
	void
updateline()
{
	int 		row;
	int 		n;

	screenalloc();		/* allocate screen buffers if size changed */

	if (Nextscreen == NULL || RedrawingDisabled)
		return;

	screenchar(NULL, 0, 0);	/* init cursor position of screenchar() */
	outstr(T_CI);				/* disable cursor */
	row = screenline(Curpos.lnum, Cline_row, (int)Rows - 1);

	outstr(T_CV);				/* enable cursor again */

	if (row == Rows)			/* line too long for screen */
		updateScreen(VALID_TO_CURSCHAR);
	else
	{
		n = row - Cline_row;
		if (n != Cline_size)		/* line changed size */
		{
			if (n < Cline_size) 	/* got smaller: delete lines */
					s_del(row, Cline_size - n, FALSE);
			else					/* got bigger: insert lines */
					s_ins(Cline_row + Cline_size, n - Cline_size, FALSE);

			updateScreen(VALID_TO_CURSCHAR);
		}
	}
}

/*
 * updateScreen()
 *
 * Based on the current value of Topline, transfer a screenfull of stuff from
 * Filemem to Nextscreen, and update Botline.
 */

	void
updateScreen(type)
	int 			type;
{
	register int	row;
	register int	endrow;
	linenr_t		lnum;
	linenr_t		lastline = 0; /* only valid if endrow != Rows -1 */
	int				done;		/* if TRUE, we hit the end of the file */
	int				didline;	/* if TRUE, we finished the last line */
	int 			srow = 0;	/* starting row of the current line */
	int 			idx;
	int 			i;
	long 			j;
	static int		postponed_not_valid = FALSE;
	register u_char *screenp;
	register u_char *attribp;

	screen_msg(FALSE, NULL);
	screenalloc();		/* allocate screen buffers if size changed */

	if (Nextscreen == NULL)
		return;

	if (type == CLEAR)		/* first clear screen */
	{
		screenclear();
		type = NOT_VALID;
	}
	if (type == CURSUPD)	/* update cursor and then redraw */
	{
		NumLineSizes = 0;
		cursupdate();		/* will call updateScreen(VALID) */
		return;
	}
	if (NumLineSizes == 0)
		type = NOT_VALID;

 	if (RedrawingDisabled)
	{
		if (type == NOT_VALID)
			postponed_not_valid = TRUE;		/* use NOT_VALID next time */
		return;
	}

	if (postponed_not_valid)
	{
		type = NOT_VALID;
		postponed_not_valid = FALSE;
	}

/* return if there is nothing to do */
	if ((type == VALID && Topline == LineNumbers[0]) ||
			(type == INVERTED && oldCurpos.lnum == Curpos.lnum &&
					oldCurpos.col == Curpos.col))
		return;

	if (type == NOT_VALID)
	{
		redraw_msg = TRUE;
		NumLineSizes = 0;
	}

	idx = 0;
	row = 0;
	lnum = Topline;
	outstr(T_CI);				/* disable cursor */

	/* The number of rows shown is Rows-1. */
	/* The default last row is the status/command line. */
	endrow = Rows - 1;

	if (type == VALID || type == VALID_TO_CURSCHAR)
	{
		/*
		 * We handle two special cases:
		 * 1: we are off the top of the screen by a few lines: scroll down
		 * 2: Topline is below LineNumbers[0]: may scroll up
		 */
		if (Topline < LineNumbers[0])	/* may scroll down */
		{
			j = LineNumbers[0] - Topline;
			if (j < Rows - 3)				/* not too far off */
			{
				lastline = LineNumbers[0] - 1;
				i = plines_m(Topline, lastline);
				if (i < Rows - 3)		/* less than a screen off */
				{
					/*
					 * Try to insert the correct number of lines.
					 * This may fail and the screen may have been cleared.
					 */
					if (s_ins(0, i, FALSE) && NumLineSizes)
					{
						endrow = i;

						if ((NumLineSizes += j) > Rows - 1)
							NumLineSizes = Rows - 1;
						for (idx = NumLineSizes; idx - j >= 0; idx--)
						{
							LineNumbers[idx] = LineNumbers[idx - j];
							LineSizes[idx] = LineSizes[idx - j];
						}
						idx = 0;
					}
				}
				else		/* far off: clearing the screen is faster */
					screenclear();
			}
			else		/* far off: clearing the screen is faster */
				screenclear();
		}
		else							/* may scroll up */
		{
			j = -1;
			for (i = 0; i < NumLineSizes; i++) /* try to find Topline in LineNumbers[] */
			{
				if (LineNumbers[i] == Topline)
				{
					j = i;
					break;
				}
				row += LineSizes[i];
			}
			if (j == -1)	/* Topline is not in LineNumbers */
			{
				row = 0;
				screenclear();   /* far off: clearing the screen is faster */
			}
			else
			{
				/*
				 * Try to delete the correct number of lines.
				 * Topline is at LineNumbers[i].
				 */
				if ((row == 0 || s_del(0, row, FALSE)) && NumLineSizes)
				{
					srow = row;
					row = 0;
					for (;;)
					{
						if (type == VALID_TO_CURSCHAR && lnum == Curpos.lnum)
								break;
						if (row + srow + LineSizes[j] >= Rows - 1)
								break;
						LineSizes[idx] = LineSizes[j];
						LineNumbers[idx] = lnum++;

						row += LineSizes[idx++];
						if (++j >= NumLineSizes)
							break;
					}
					NumLineSizes = idx;
				}
				else
					row = 0;		/* update all lines */
			}
		}
		if (endrow == Rows - 1 && idx == 0) 	/* no scrolling */
				NumLineSizes = 0;
	}

	done = didline = FALSE;
	screenchar(NULL, 0, 0);	/* init cursor position of screenchar() */

	if (Quote.lnum)				/* check if we are updating the inverted part */
	{
		linenr_t	from, to;

	/* find the line numbers that need to be updated */
		if (Curpos.lnum < oldCurpos.lnum)
		{
			from = Curpos.lnum;
			to = oldCurpos.lnum;
		}
		else
		{
			from = oldCurpos.lnum;
			to = Curpos.lnum;
		}
	/* if in block mode and changed column or Curswant: update all lines */
		if (Quote_block && (Curpos.col != oldCurpos.col || Curswant != oldCurswant))
		{
			if (from > Quote.lnum)
				from = Quote.lnum;
			if (to < Quote.lnum)
				to = Quote.lnum;
		}

		if (from < Topline)
			from = Topline;
		if (to >= Botline)
			to = Botline - 1;

	/* find the minimal part to be updated */
		if (type == INVERTED)
		{
			while (lnum < from)						/* find start */
			{
				row += LineSizes[idx++];
				++lnum;
			}
			srow = row;
			for (j = idx; j < NumLineSizes; ++j)	/* find end */
			{
				if (LineNumbers[j] == to + 1)
				{
					endrow = srow;
					break;
				}
				srow += LineSizes[j];
			}
			oldCurpos = Curpos;
			oldCurswant = Curswant;
		}
	/* if we update the lines between from and to set oldCurpos */
		else if (lnum <= from && (endrow == Rows - 1 || lastline >= to))
		{
			oldCurpos = Curpos;
			oldCurswant = Curswant;
		}
	}

	/*
	 * Update the screen rows from "row" to "endrow".
	 * Start at line "lnum" which is at LineNumbers[idx].
	 */
	for (;;)
	{
			if (lnum > line_count)		/* hit the end of the file */
			{
				done = TRUE;
				break;
			}
			srow = row;
			row = screenline(lnum, srow, endrow);
			if (row > endrow)	/* past end of screen */
			{
				LineSizes[idx] = plines(lnum);	/* we may need the size of that */
				LineNumbers[idx++] = lnum;		/* too long line later on */
				break;
			}

			LineSizes[idx] = row - srow;
			LineNumbers[idx++] = lnum;
			if (++lnum > line_count)
			{
				done = TRUE;
				break;
			}

			if (row == endrow)
			{
				didline = TRUE;
				break;
			}
	}
	if (idx > NumLineSizes)
		NumLineSizes = idx;

	/* Do we have to do off the top of the screen processing ? */
	if (endrow != Rows - 1)
	{
		row = 0;
		for (idx = 0; idx < NumLineSizes && row < (Rows - 1); idx++)
			row += LineSizes[idx];

		if (row < (Rows - 1))
		{
			done = TRUE;
		}
		else if (row > (Rows - 1))		/* Need to blank out the last line */
		{
			lnum = LineNumbers[idx - 1];
			srow = row - LineSizes[idx - 1];
			didline = FALSE;
		}
		else
		{
			lnum = LineNumbers[idx - 1] + 1;
			didline = TRUE;
		}
	}

	emptyrows = 0;
	/*
	 * If we didn't hit the end of the file, and we didn't finish the last
	 * line we were working on, then the line didn't fit.
	 */
	if (!done && !didline)
	{
		if (lnum == Topline)
		{
			/*
			 * Single line that does not fit!
			 * Fill last line with '@' characters.
			 */
			screenp = LinePointers[Rows - 2];
			attribp = ScreenAttr(screenp);
			for (i = 0; i < Columns; ++i)
			{
				if (*screenp != '@' || *attribp)
				{
					*screenp = '@';
					*attribp = AM_None;
					screenchar(screenp, (int)(Rows - 2), i);
				}
				++screenp;
				++attribp;
			}
			Botline = lnum + 1;
		}
		else
		{
			/*
			 * Clear the rest of the screen and mark the unused lines.
			 */
			screenfill(srow, '@');
			Botline = lnum;
		}
	}
	else
	{
		/* make sure the rest of the screen is blank */
		/* put '~'s on rows that aren't part of the file. */
		screenfill(row, '~');
		emptyrows = Rows - row - 1;

		if (done)				/* we hit the end of the file */
			Botline = line_count + 1;
		else
			Botline = lnum;
	}

	if (redraw_msg)
	{
		showmode();
		redraw_msg = FALSE;
	}

	outstr(T_CV);				/* enable cursor again */
}


/*
 * Move line "lnum" to the screen.
 * Start at row "startrow", stop when "endrow" is reached.
 * Return the number of last row the line occupies.
 */

#ifdef KANJI
static int		kanji;
static u_char	kcode = NUL, *kin = NULL, *kout = NULL;
#endif

	static int
screenline(lnum, startrow, endrow)
		linenr_t		lnum;
		int 			startrow;
		int 			endrow;
{
	register u_char  *screenp;
	register u_char  *attribp;
	register u_char   c;
	register int	col;				/* visual column on screen */
	register int	vcol;				/* visual column for tabs */
	register int	row;
	register u_char *ptr;
	char			extra[16];
	char			*p_extra;
	int 			n_extra;

	int				invert;

	int				fromcol, tocol;		/* start/end of inverting */
#ifdef FEP
	int				uline;

	int				kufrom, kuto;		/* start/end of underline */
	int				krfrom, krto;		/* start/end of inverting */
#endif /* FEP */
	int				temp;
	FPOS			*top, *bot;

#ifdef KANJI

	if (kcode != KANJI_DISP) 
		kcode = NUL;
#endif /* KANJI */
	row = startrow;
	col = 0;
	vcol = 0;
	invert = FALSE;
	fromcol = tocol = -10;
#ifdef FEP
	uline = FALSE;
#endif
	ptr = (u_char *)nr2ptr(lnum);
	canopt = TRUE;
	if (Quote.lnum)					/* quoting active */
	{
		if (ltoreq(Curpos, Quote))		/* Quote is after Curpos */
		{
			top = &Curpos;
			bot = &Quote;
		}
		else							/* Quote is before Curpos */
		{
			top = &Quote;
			bot = &Curpos;
		}
		if (Quote_block)						/* block mode */
		{
			if (lnum >= top->lnum && lnum <= bot->lnum)
			{
				fromcol = getvcol(top, 2);
				temp = getvcol(bot, 2);
				if (temp < fromcol)
					fromcol = temp;

				if (Curswant == 29999)
					tocol = 29999;
				else
				{
					tocol = getvcol(top, 3);
					temp = getvcol(bot, 3);
					if (temp > tocol)
						tocol = temp;
					++tocol;
				}
			}
		}
		else							/* non-block mode */
		{
			if (lnum >= top->lnum && lnum <= bot->lnum)
				{
					fromcol = 0;
					tocol = Columns;
				}
			if (lnum == top->lnum)
					fromcol = getvcol(top, 2);
			if (lnum == bot->lnum)
					tocol = getvcol(bot, 3) + 1;

			if (Quote.col == QUOTELINE)		/* linewise */
			{
				if (fromcol > 0)
					fromcol = 0;
				if (tocol > 0)
					tocol = QUOTELINE;
			}
		}
		/* if inverting in this line, can't optimize cursor positioning */
		if (fromcol >= 0 || tocol >= 0)
			canopt = FALSE;
	}
#ifdef FEP
	if (Kconvlnum == lnum)
	{
		FPOS	tmp;
		tmp.lnum = lnum;
		tmp.col = KconvAltStart;	krfrom = getvcol(&tmp, 2);
		tmp.col = KconvAltEnd;		krto   = getvcol(&tmp, 3);
		tmp.col = KconvStart;		kufrom = getvcol(&tmp, 2);
		tmp.col = KconvEnd;			kuto   = getvcol(&tmp, 3);
	}
	else
		krfrom = kufrom = krto = kuto = -10;
#endif /* FEP */

	screenp = LinePointers[row];
	attribp = ScreenAttr(screenp);
	if (p_nu)
	{
		sprintf(extra, "%7ld ", (long)lnum);
		p_extra = extra;
		n_extra = 8;
		vcol = -8;		/* so vcol is 0 when line number has been printed */

	}
	else
	{
		p_extra = NULL;
		n_extra = 0;
	}

#ifdef KANJI
	kanji = FALSE;
#endif /* KANJI */


	for (;;)
	{
		if (invert)
		{
#ifdef FEP
			if (vcol >= krto && vcol >= tocol)		/* stop inverting */
#else
			if (vcol >= tocol)
#endif
				invert = invert_off(invert);
		}
		else
		{
#ifdef FEP
			if (vcol >= krfrom && vcol < krto)		/* start inverting */
			{
				uline = uline_off(uline);
				kufrom = krto;
				invert = invert_on(invert);
			}
			else
#endif
			if (vcol >= fromcol && vcol < tocol)		/* start inverting */
				invert = invert_on(invert);
		}

#ifdef FEP
		if (uline)
		{
			if (vcol >= kuto)
				 uline = uline_off(uline);
		}
		else
		{
			if (vcol >= kufrom && vcol < kuto)
				 uline = uline_on(uline);
		}
#endif

		/* Get the next character to put on the screen. */
		/*
		 * The 'extra' array contains the extra stuff that is inserted to
		 * represent special characters (tabs, and other non-printable stuff.
		 * The order in the 'extra' array is reversed.
		 */

		if (n_extra > 0)
		{
			c = (u_char)*p_extra++;
			n_extra--;
		}
		else
		{
			if ((c = *ptr++) < ' ' || (c > '~' && c < 0xa0))
			{
				/*
				 * when getting a character from the file, we may have to turn it
				 * into something else on the way to putting it into 'Nextscreen'.
				 */
				if (c == TAB && !p_list)
				{
					p_extra = spaces;
					/* tab amount depends on current column */
					n_extra = (int)p_ts - vcol % (int)p_ts - 1;
					c = ' ';
				}
				else if (c == NUL && p_list)
				{
					extra[0] = NUL;
					p_extra = extra;
					n_extra = 1;
					c = '$';
				}
				else if (c != NUL)
				{
					p_extra = (char *)transchar(c);
					n_extra = charsize(c) - 1;
					c = (u_char)*p_extra++;
				}
			}
		}

		if (c == NUL)
		{
			if (invert)
			{
				if (vcol == 0)	/* invert first char of empty line */
				{
					if (*screenp != ' ' || !(*attribp & AM_Inv))
					{
						*screenp = ' ';
						*attribp = AM_Inv;
						screenchar(screenp, row, col);
					}
					++screenp;
					++attribp;
					++col;
				}
				invert = invert_off(invert);
			}
#ifdef UNIX
			/* 
			 * blank out the rest of this row
		     */
		    if (col < Columns)
			{
			  windgoto(row, col);
			  clear_line();

			  while(col < Columns)
			  {
			    *screenp ++ = ' ';
			    *attribp ++ = AM_None;
				col ++;
			  }
			}
#else
			/*
			 * could also use clear-to-end-of-line, but it is slower
			 * on an Amiga
			 */
			while (col < Columns)
			{
				if (*screenp != ' ' || *attribp)
				{
					*screenp = ' ';
					*attribp = AM_None;
					screenchar(screenp, row, col);
				}
				screenp++;
				attribp++;
				col++;
			}
#endif
			row++;
#ifdef KANJI
			kanji = kanji_off(kanji);
#endif /* KANJI */
			break;
		}
#ifdef KANJI
		if (col >= (IsKanji(c) ? Columns - 1 : Columns)) /* continuous line */
		{
			kanji = kanji_off(kanji);

			if (col == Columns - 1 &&
								(*screenp != '\\' || *attribp != AM_None))
			{
				*screenp = '\\';
				*attribp = AM_None;
				screenchar(screenp, row, col);
			}

			col = 0;
			if (++row == endrow)		/* line got too long for screen */
			{
				++row;
				break;
			}
			screenp = LinePointers[row];
			attribp = ScreenAttr(screenp);
#ifdef FEP
			uline  = uline_off(uline);
#endif
			screen_msg(FALSE, NULL);
		}
#else /* KANJI */
		if (col >= Columns)
		{
			col = 0;
			if (++row == endrow)		/* line got too long for screen */
			{
				++row;
				break;
			}
			screenp = LinePointers[row];
			attribp = ScreenAttr(screenp);
			screen_msg(FALSE, NULL);
		}
#endif /* KANJI */

		/* store the character in Nextscreen */

#ifdef KANJI
		if (IsKanji(c) && KANJI_DISP != KCODE_NONE)
		{	u_char c1, c2;
			int attr;

			c1 = c;
			c2 = *ptr++;
			attr = *attribp;

			if (*screenp != c1 || *(screenp + 1) != c2
				|| !(attr & AM_K1)
				|| (invert ? !(attr & AM_Inv) : (attr & AM_Inv))
#ifdef FEP
		  		|| (uline ?  !(attr & AM_UL) :  (attr & AM_UL))
#endif
				)
			{
				kanji = kanji_on(kanji);

				if (invert)
				{
					 *attribp++ = AM_Inv | AM_K1;
					 *attribp++ = AM_Inv;
				}
#ifdef FEP
				else if (uline)
				{
					 *attribp++ = AM_UL | AM_K1;
					 *attribp++ = AM_UL;
				}
#endif
				else
				{
					 *attribp++ = AM_K1;
					 *attribp++ = AM_None;
				}
				*screenp++ = c1;
				*screenp++ = c2;
				screenchar(screenp - 2, row, col);
				col  += 2;
				vcol += 2;
			}
			else
			{
				invert = invert_off(invert);
#ifdef FEP
				uline  = uline_off(uline);
#endif
				attribp += 2;
				screenp += 2;
				vcol    += 2;
				col     += 2;
			}
		}
		else
#endif /* KANJI */
		{
			int attr;

			attr = *attribp;
			if (*screenp != c
				|| (invert ? !(attr & AM_Inv) : (attr & AM_Inv))
#ifdef FEP
				|| (uline ?  !(attr & AM_UL) :  (attr & AM_UL))
#endif
			   )
			{
#ifdef KANJI
				kanji = kanji_off(kanji);
#endif /* KANJI */

				*screenp = c;
				if      (invert) *attribp = AM_Inv;
#ifdef FEP
				else if (uline)  *attribp = AM_UL;
#endif
				else	*attribp = AM_None;
				screenchar(screenp, row, col);
			}
			else
			{
				invert = invert_off(invert);
#ifdef FEP
				uline  = uline_off(uline);
#endif
			}

			screenp ++;
			attribp ++;
			vcol ++;
			col  ++;
		}
	}

#ifdef KANJI
	kanji = kanji_off(kanji);
#endif
#ifdef FEP
	uline = uline_off(uline);
#endif
	invert = invert_off(invert);

	return (row);
}

/*
 * put character '*p' on the screen at position 'row' and 'col'
 */
	static void
screenchar(p, row, col)
		u_char	*p;
		int 	row;
		int 	col;
{
	static int	oldrow, oldcol;		/* old cursor position */
	int			c;

	if (p == NULL)					/* initialize cursor position */
	{
		oldrow = oldcol = -1;
		return;
	}
	if (oldcol != col || oldrow != row)
	{
		/*
		 * If we're on the same row (which happens a lot!), try to
		 * avoid a windgoto().
		 * If we are only a few characters off, output the
		 * characters. That is faster than cursor positioning.
		 * This can't be used when inverting (a part of) the line.
		 */
		if (oldrow == row && oldcol < col)
		{
			register int i;
			i = col - oldcol;
#ifdef KANJI
			if (i <= 10 && canopt && !IsKanji(*(p - i)))
			{
				char	c2;

				while (i)
				{
					c = *(p - i--);
					if (IsKanji(c))
					{
					    kanji = kanji_on(kanji);
						c2 = *(p - i--);
						kanjito(&c, &c2, KANJI_DISP);
						outchar(c);
						outchar(c2);
					}
					else
					{
					    kanji = kanji_off(kanji);
						outchar(c);
					}
				}
			    kanji = IsKanji(*p) ?  kanji_on(kanji) : kanji_off(kanji);
			}
			else
#else /* KANJI */
			if (i <= 4 && canopt)
			{
				while (i)
				{
					c = *(p - i--);
					outchar(c);
				}
			}
			else
#endif /* KANJI */
			{
#ifdef KANJI
				kanji = kanji_off(kanji);
#endif
				if (T_CRI && *T_CRI)	/* use tgoto interface! jw */
					outstr(tgoto(T_CRI, 0, i));
				else
					windgoto(row, col);
			}
			
			oldcol = col;
		}
		else
		{
#ifdef KANJI
			kanji = kanji_off(kanji);
#endif
			windgoto(oldrow = row, oldcol = col);
		}
	}
#ifdef KANJI
	if (IsKanji(*p))
	{
		u_char c1, c2;
		c1 = *p++;
		c2 = *p;
		kanji = kanji_on(kanji);
		kanjito(&c1, &c2, KANJI_DISP);
		outchar(c1);
		outchar(c2);
		oldcol++;
	}
	else
		outchar(*p);
#else
	outchar(*p);
#endif

	oldcol++;
}

/*
 * Fill the screen at 'srow' with character 'c' followed by blanks.
 */
	static void
screenfill(srow, c)
		int 	srow;
		int		c;
{
		register int row;
		register int col;
		register u_char *screenp;
		register u_char *attribp;

		for (row = srow; row < (Rows - 1); ++row)
		{
			screenp = LinePointers[row];
			attribp = ScreenAttr(screenp);

			if (*screenp != c || *attribp)
			{
				*screenp = c;
				*attribp = AM_None;
				screenchar(screenp, row, 0);
			}
			++attribp;
			++screenp;
			for (col = 1; col < Columns; ++col)
			{
				if (*screenp != ' ' || *attribp)
				{
					*screenp = ' ';
					*attribp = AM_None;
					screenchar(screenp, row, col);
				}
				++attribp;
				++screenp;
			}
		}
}

/*
 * compute Botline. Can be called after Topline or Rows changed.
 */
	void
comp_Botline()
{
	linenr_t	lnum;
	int			done = 0;

	for (lnum = Topline; lnum <= line_count; ++lnum)
	{
		if ((done += plines(lnum)) >= Rows)
			break;
	}
	Botline = lnum;		/* Botline is the line that is just below the window */
}

/*
 * prt_line() - print the given line
 * returns the number of characters written.
 */
	int
prt_line(s)
	char		   *s;
{
	register int	si = 0;
	register char	c;
	register int	col = 0;

	char			extra[16];
	int 			n_extra = 0;
	int 			n;

	for (;;)
	{
		if (n_extra > 0)
			c = extra[--n_extra];
		else {
			c = s[si++];
			if (c == TAB && !p_list)
			{
				strcpy(extra, "                ");
				/* tab amount depends on current column */
				n_extra = (p_ts - 1) - col % p_ts;
				c = ' ';
			}
			else if (c == NUL && p_list)
			{
				extra[0] = NUL;
				n_extra = 1;
				c = '$';
			}
			else if (c != NUL && (n = charsize(c)) > 1)
			{
				char			 *p;

				n_extra = 0;
				p = transchar(c);
				/* copy 'ch-str'ing into 'extra' in reverse */
				while (n > 1)
					extra[n_extra++] = p[--n];
				c = p[0];
			}
		}

		if (c == NUL)
			break;

		outchar(c);
		col++;
	}
	return col;
}

	void
screenalloc()
{
	static int		old_Rows = 0;
	static int		old_Columns = 0;
	register int	i;

	/*
	 * Allocation of the sceen buffers is done only when the size changes
	 */
	if (Nextscreen != NULL && Rows == old_Rows && Columns == old_Columns)
		return;

	old_Rows = Rows;
	old_Columns = Columns;

	/*
	 * If we're changing the size of the screen, free the old arrays
	 */
	if (Nextscreen != NULL)
		free((char *)Nextscreen);
	if (Nextattrib != NULL)
		free((char *)Nextattrib);
	if (LinePointers != NULL)
		free((char *)LinePointers);
	if (LineNumbers != NULL)
		free((char *) LineNumbers);
	if (LineSizes != NULL)
		free(LineSizes);

	Nextscreen = (u_char *)malloc((size_t) (Rows * Columns));
	Nextattrib = (u_char *)malloc((size_t) (Rows * Columns));
	LineNumbers = (linenr_t *) malloc((size_t) (Rows * sizeof(linenr_t)));
	LineSizes = (u_char *)malloc((size_t) Rows);
	LinePointers = (u_char **)malloc(sizeof(u_char *) * Rows);

	if (Nextscreen == NULL || Nextattrib == NULL 
			|| LineNumbers == NULL || LineSizes == NULL || LinePointers == NULL)
	{
		emsg(e_outofmem);
		if (Nextscreen != NULL)
			{ free((char *)Nextscreen);
			  free((char *)Nextattrib);
			}
		Nextscreen = Nextattrib = NULL;
	}
	else
	{
		for (i = 0; i < Rows; ++i)
				LinePointers[i] = Nextscreen + i * Columns;
	}

	screenclear();
}

	void
screenclear()
{
	register u_char  *np;
	register u_char  *ap;
	register u_char  *end;

	if (starting || Nextscreen == NULL)
		return;

	outstr(T_ED);				/* clear the display */

	np = Nextscreen;
	ap = Nextattrib;
	end = Nextscreen + Rows * Columns;

	/* blank out Nextscreen */
	while (np != end)
	{
		*ap++ = NUL;
		*np++ = ' ';
	}

	/* clear screen info */
	NumLineSizes = 0;
}

	void
cursupdate()
{
	linenr_t		p;
	long 			nlines;
	int 			i;
	int 			temp;

	screenalloc();		/* allocate screen buffers if size changed */

	if (Nextscreen == NULL)
		return;

	if (Curpos.lnum > line_count)
		Curpos.lnum = line_count;
	if (bufempty()) 			/* special case - file is empty */
	{
		Topline = 1;
		Curpos.lnum = 1;
		Curpos.col = 0;
		LineSizes[0] = 1;
		for (i = 1; i < Rows; i++)
			LineSizes[i] = 0;
		if (NumLineSizes == 0)		/* don't know about screen contents */
			updateScreen(NOT_VALID);
		NumLineSizes = 1;
	}
	else if (Curpos.lnum < Topline)
	{
		/*
		 * If the cursor is above the top of the screen, scroll the screen to
		 * put it at the top of the screen.
		 * If we weren't very close to begin with, we scroll more, so that
		 * the line is close to the middle.
		 */
		temp = Rows / 3;
		if (Topline - Curpos.lnum >= temp)
		{
			p = Curpos.lnum;
			for (i = 0; i < temp && p > 1; i += plines(--p))
				;
			Topline = p;
		}
		else if (p_sj > 1)		/* scroll at least p_sj lines */
		{
			for (i = 0; i < p_sj && Topline > 1; i += plines(--Topline))
				;

		}
		if (Topline > Curpos.lnum)
			Topline = Curpos.lnum;
		updateScreen(VALID);
	}
	else if (Curpos.lnum >= Botline)
	{
		nlines = Curpos.lnum - Botline + 1;
		/*
		 * compute the number of lines at the top which have the same or more
		 * rows than the rows of the lines below the bottom
		 */
		if (nlines <= Rows)
		{
				/* get the number or rows to scroll minus the number of
								free '~' rows */
			temp = plines_m(Botline, Curpos.lnum) - emptyrows;
			if (temp <= 0)
				nlines = 0;
			else
			{
					/* scroll minimal number of lines */
				if (temp < p_sj)
					temp = p_sj;
				for (i = 0, p = Topline; i < temp && p < Botline; ++p)
					i += plines(p);
				nlines = p - Topline;
			}
		}

		/*
		 * Scroll up if the cursor is off the bottom of the screen a bit.
		 * Otherwise put it at 2/3 of the screen.
		 */
		if (nlines > Rows / 3 && nlines > p_sj)
		{
			p = Curpos.lnum;
			temp = (2 * Rows) / 3;
			nlines = 0;
			i = 0;
			do				/* this loop could win a contest ... */
				i += plines(p);
			while (i < temp && (nlines = 1) != 0 && --p != 0);
			Topline = p + nlines;
		}
		else
			scrollup(nlines);
		updateScreen(VALID);
	}
	else if (NumLineSizes == 0)		/* don't know about screen contents */
		updateScreen(NOT_VALID);
	Cursrow = Curscol = Cursvcol = i = 0;
	for (p = Topline; p != Curpos.lnum; ++p)
		if (RedrawingDisabled)		/* LineSizes[] invalid */
			Cursrow += plines(p);
		else
			Cursrow += LineSizes[i++];

	Cline_row = Cursrow;
	if (!RedrawingDisabled && i > NumLineSizes)
								/* Should only happen with a line that is too */
								/* long to fit on the last screen line. */
		Cline_size = 0;
	else
	{
		if (RedrawingDisabled)      /* LineSizes[] invalid */
		    Cline_size = plines(Curpos.lnum);
        else
			Cline_size = LineSizes[i];
		curs_columns();		/* compute Cursvcol and Curscol */
	}

	if (set_want_col)
	{
		Curswant = Cursvcol;
		set_want_col = FALSE;
	}
	showruler(0);
}

/*
 * compute Curscol and Cursvcol
 */
	void
curs_columns()
{	
#ifdef KANJI
	u_char *ptr, c;
	int  col;

	Cursvcol = 0;
	if (p_nu)
		Curscol = 8;
	else
		Curscol = 0;
	Cursrow = Cline_row;

	ptr = (u_char *)nr2ptr(Curpos.lnum);

	if (!ptr)
	{
		int lnum, col;
		lnum = Curpos.lnum;
		col  = Curpos.col;

		Curpos.lnum = Curpos.col = 1;
		updateScreen(NOT_VALID);
		smsg("JVIM zombie: curpos(line:%s, col:%s)", lnum, col);
		sleep(1);
		return;
	}

	col = 0;
	while(col < Curpos.col)
	{
		int lineend;

		c = *ptr;
		if (IsKanji(c))
		{
			col += 2;
			ptr += 2;
			Cursvcol += 2;
			Curscol  += 2;
			lineend = Curscol >= Columns - 1;
		}
		else
		{
			int cw;
			cw  = chartabsize(c, Curscol);
			col ++;
			ptr ++;
			Cursvcol += cw;
			Curscol  += cw;
			lineend = Curscol >= Columns;
		}

		if (lineend)
		{
		    Cursrow++;
			Curscol = 0;
		}
	}
	if (!IsKanji(c = *ptr))
	  if (c == TAB && State == NORMAL && !p_list)
	  { int cw;

	    cw = chartabsize(c, Curscol) - 1;
		Cursvcol += cw;
		Curscol  += cw;
	  }
#else
	Cursvcol = getvcol(&Curpos, 1);
	Curscol = Cursvcol;
	if (p_nu)
		Curscol += 8;

	Cursrow = Cline_row;

	while (Curscol >= Columns)
	{
		Curscol -= Columns;
		Cursrow++;
	}
#endif

	if (Cursrow > Rows - 2)		/* Cursor past end of screen */
		Cursrow = Rows - 2;		/* happens with line that does not fit on screen */
}

/*
 * get virtual column number of pos
 * type = 1: where the cursor is on this character
 * type = 2: on the first position of this character
 * type = 3: on the last position of this character
 */
	int
getvcol(pos, type)
	FPOS	*pos;
	int		type;
{
	int				col;
	int				vcol;
	u_char		   *ptr;
	int 			incr;
	u_char			c;

	vcol = 0;
	ptr = (u_char *)nr2ptr(pos->lnum);
	for (col = pos->col; col >= 0; --col)
	{
		c = *ptr++;
		if (c == NUL)		/* make sure we don't go past the end of the line */
			break;

		/* A tab gets expanded, depending on the current column */
		incr = chartabsize(c, vcol);

		if (col == 0)		/* character at pos.col */
		{
			if (type == 3 || (type == 1 && c == TAB && State == NORMAL && !p_list))
				--incr;
			else
				break;
		}
		vcol += incr;
	}
	return vcol;
}

	void
scrolldown(nlines)
	long	nlines;
{
	register long	done = 0;	/* total # of physical lines done */

	/* Scroll up 'nlines' lines. */
	while (nlines--)
	{
		if (Topline == 1)
			break;
		done += plines(--Topline);
	}
	/*
	 * Compute the row number of the last row of the cursor line
	 * and move it onto the screen.
	 */
	Cursrow += done + plines(Curpos.lnum) - 1 - Cursvcol / Columns;
	while (Cursrow >= Rows - 1 && Curpos.lnum > 1)
		Cursrow -= plines(Curpos.lnum--);
}

	void
scrollup(nlines)
	long	nlines;
{
#ifdef NEVER
	register long	done = 0;	/* total # of physical lines done */

	/* Scroll down 'nlines' lines. */
	while (nlines--)
	{
		if (Topline == line_count)
			break;
		done += plines(Topline);
		if (Curpos.lnum == Topline)
			++Curpos.lnum;
		++Topline;
	}
	s_del(0, done, TRUE);
#endif
	Topline += nlines;
	if (Topline > line_count)
		Topline = line_count;
	if (Curpos.lnum < Topline)
		Curpos.lnum = Topline;
}

/*
 * The rest of the routines in this file perform screen manipulations. The
 * given operation is performed physically on the screen. The corresponding
 * change is also made to the internal screen image. In this way, the editor
 * anticipates the effect of editing changes on the appearance of the screen.
 * That way, when we call screenupdate a complete redraw isn't usually
 * necessary. Another advantage is that we can keep adding code to anticipate
 * screen changes, and in the meantime, everything still works.
 */

/*
 * s_ins(row, nlines, invalid) - insert 'nlines' lines at 'row'
 * if 'invalid' is TRUE the LineNumbers[] is invalidated.
 * Returns 0 if the lines are not inserted, 1 for success.
 */
	int
s_ins(row, nlines, invalid)
	int 		row;
	int 		nlines;
	int			invalid;
{
	int 		i;
	int 		j;
	u_char		*temp;

	screenalloc();		/* allocate screen buffers if size changed */

	if (Nextscreen == NULL)
		return 0;

	if (invalid)
		NumLineSizes = 0;

	if (nlines > (Rows - 1 - row))
		nlines = Rows - 1 - row;

	if (RedrawingDisabled || nlines <= 0 ||
						((T_CIL == NULL || *T_CIL == NUL) &&
						(T_IL == NULL || *T_IL == NUL) &&
						(T_SR == NULL || *T_SR == NUL || row != 0)))
		return 0;
	
	if (Rows - nlines < 5)	/* only a few lines left: redraw is faster */
	{
		screenclear();		/* will set NumLineSizes to 0 */
		return 0;
	}

	/*
	 * It "looks" better if we do all the inserts at once
	 */
    if (T_CIL && *T_CIL) 
    {
        windgoto(row, 0);
		if (nlines == 1 && T_IL && *T_IL)
			outstr(T_IL);
		else
			outstr(tgoto(T_CIL, 0, nlines));
    }
    else
    {
        for (i = 0; i < nlines; i++) 
        {
            if (i == 0 || row != 0)
				windgoto(row, 0);
			if (T_IL && *T_IL)
				outstr(T_IL);
			else
				outstr(T_SR);
        }
    }
	windgoto((int)Rows - 1, 0);		/* delete any garbage that may have */
	clear_line();					/* been shifted to the bottom line */
	redraw_msg = TRUE;

	/*
	 * Now shift LinePointers nlines down to reflect the inserted lines.
	 * Clear the inserted lines.
	 */
	for (i = 0; i < nlines; ++i)
	{
		j = Rows - 2 - i;
		temp = LinePointers[j];
		while ((j -= nlines) >= row)
				LinePointers[j + nlines] = LinePointers[j];
		LinePointers[j + nlines] = temp;
		memset(temp, ' ', (size_t)Columns);
	}
	return 1;
}

/*
 * s_del(row, nlines, invalid) - delete 'nlines' lines at 'row'
 * If 'invalid' is TRUE LineNumbers[] is invalidated.
 * Return 1 for success, 0 if the lines are not deleted.
 */
	int
s_del(row, nlines, invalid)
	int 			row;
	int 			nlines;
	int			invalid;
{
	int 			j;
	int 			i;
	u_char		*temp;

	screenalloc();		/* allocate screen buffers if size changed */

	if (Nextscreen == NULL)
		return 0;

	if (invalid)
		NumLineSizes = 0;

	if (nlines > (Rows - 1 - row))
		nlines = Rows - 1 - row;

	if (RedrawingDisabled || nlines <= 0 ||
				((T_DL == NULL || *T_DL == NUL) &&
				(T_CDL == NULL || *T_CDL == NUL) &&
				row != 0))
		return 0;

	if (Rows - nlines < 5)	/* only a few lines left: redraw is faster */
	{
		screenclear();		/* will set NumLineSizes to 0 */
		return 0;
	}

	windgoto((int)Rows - 1, 0);		/* delete any garbage that may be */
	clear_line();					/* on the bottom line */
	redraw_msg = TRUE;

	/* delete the lines */
	if (T_CDL && *T_CDL) 
	{
		windgoto(row, 0);
		if (nlines == 1 && T_DL && *T_DL)
			outstr(T_DL);
		else
			outstr(tgoto(T_CDL, 0, nlines));
	} 
	else
	{
		if (row == 0)
		{
			for (i = 0; i < nlines; i++) 
				outchar('\n');
		}
		else
		{
			for (i = 0; i < nlines; i++) 
			{
				windgoto(row, 0);
				outstr(T_DL);           /* delete a line */
			}
		}
	}

	/*
	 * Now shift LinePointers nlines up to reflect the deleted lines.
	 * Clear the deleted lines.
	 */
	for (i = 0; i < nlines; ++i)
	{
		j = row + i;
		temp = LinePointers[j];
		while ((j += nlines) < Rows - 1)
				LinePointers[j - nlines] = LinePointers[j];
		LinePointers[j - nlines] = temp;
		memset(temp, ' ', (size_t)Columns);
	}
	return 1;
}

	void
showmode()
{
		if ((p_mo && (State == INSERT || State == REPLACE)) || Recording)
		{
				gotocmdline(TRUE, NUL);
				if (p_mo)
				{
#ifdef ONEW
					if (!KanjiInput)
					{
						if (State == INSERT)
							outstrn("[INS]");
						if (State == REPLACE)
							outstrn("[REP]");
					}
#else /* ONEW */
					if (State == INSERT)
						outstrn("-- INSERT --");
					if (State == REPLACE)
						outstrn("-- REPLACE --");
#endif /* ONEW */
				}
				if (!p_fm && Recording)
					outstrn("recording");
		}
		showruler(1);
}

/*
 * delete mode message
 */
	void
delmode()
{
	if (Recording)
		msg("recording");
	else
		msg("");
}

/*
 * if ruler option is set: show current cursor position
 * if always is FALSE, only print if position has changed
 */
	void
showruler(always)
	int		always;
{
	static linenr_t	oldlnum = 0;
	static colnr_t	oldcol = 0;
	static int		oldlen = 0;
	int				newlen;
	char			buffer[20];

	if (p_ru && (redraw_msg || always || Curpos.lnum != oldlnum || Cursvcol != oldcol))
	{
		windgoto((int)Rows - 1, (int)Columns - 22);
		/*
		 * Some sprintfs return the lenght, some return a pointer.
		 * To avoid portability problems we use strlen here.
		 */
		sprintf(buffer, "%ld,%d", Curpos.lnum, Cursvcol + 1);
		newlen = strlen(buffer);
		outstrn(buffer);
		while (newlen < oldlen)
		{
			outchar(' ');
			--oldlen;
		}
		oldlen = newlen;
		oldlnum = Curpos.lnum;
		oldcol = Cursvcol;
		redraw_msg = FALSE;
	}
}

/*
 * Clear a line. The cursor must be at the first char of the line.
 */
	void
clear_line()
{
	register int i;

	if (T_EL != NULL && *T_EL != NUL)
		outstr(T_EL);
	else
		for (i = 1; i < Columns; ++i)
			outchar(' ');
}

/*
 * Switch Attribute
 */
 	static int
invert_on(invert)
	int invert;
{
	if (!invert)
	{
		outstr(T_TI);
		invert = TRUE;
	}
	return invert;
}

 	static int
invert_off(invert)
	int invert;
{
	if (invert)
	{
		outstr(T_TP);
		invert = FALSE;
	}
	return invert;
}

#ifdef KANJI
 	static int
kanji_on(kanji)
	int kanji;
{
	if (!kcode)
	{
		kin = kanjiin(kcode = KANJI_DISP);
		kout= kanjiout(KANJI_DISP);
	}

	if (!kanji)
	{
		if (kin) outstr(kin);
		kanji = TRUE;
	}
	return kanji;
}

 	static int
kanji_off(kanji)
	int kanji;
{
	if (!kcode)
	{
		kin = kanjiin(kcode = KANJI_DISP);
		kout= kanjiout(KANJI_DISP);
	}

	if (kanji)
	{
		if (kout) outstr(kout);
		kanji = FALSE;
	}
	return kanji;
}
#endif /* KANJI */

#ifdef FEP
 	static int
uline_on(uline)
	int uline;
{
	if (!uline)
	{
		outstr(T_US);
		uline = TRUE;
	}
	return uline;
}

 	static int
uline_off(uline)
	int uline;
{
	if (uline)
	{
		outstr(T_UE);
		uline = FALSE;
	}
	return uline;
}
#endif /* FEP */

#define MAXMSG	80
screen_msg(rev, msg)
	u_char *msg;
{
	static int srow, scol, slen = 0;
	static u_char orgscr[MAXMSG + 1], orgatr[MAXMSG + 1];

	int ncol, nrow, nlen;

	u_char *mp, *screenp, *attribp, attr;
	u_char *osp, *oap;

	if (!LinePointers)
	{
		if (msg)
		{
			if (rev) outstr(T_TI);
			outstr(msg);
			if (rev) outstr(T_TP);
		}
		return;
	}

	/* do not display the same message */
	if (msg && slen)
	{
		u_char *msgp;

		screenp = LinePointers[srow] + scol;
		attribp = ScreenAttr(screenp);

		for(msgp = msg; msgp - msg < slen; msgp++, screenp++)
		  if (*msgp != *screenp) break;
		if (msgp - msg >= slen) return;
	}

	/* remove old message */
	if (slen)
	{
		screenp = LinePointers[srow];
		attribp = ScreenAttr(screenp);

		screenp += scol;
		attribp += scol;
		osp = orgscr;
		oap = orgatr;

#ifdef KANJI
		kanji = kanji_off(kanji);
#endif
		windgoto(srow, scol);
		for(; slen > 0; slen--)
		{
			char c1, c2;
			c1 = 
			*screenp++ = *osp++;
			*attribp++ = *oap++;

#ifdef KANJI
			if (IsKanji(c1))
			{
				slen --;
				c2 = 
				*screenp++ = *osp++;
				*attribp++ = *oap++;
				kanji = kanji_on(kanji);
				kanjito(&c1, &c2, KANJI_DISP);
				outchar(c1);
				outchar(c2);
			}
			else
			{
				kanji = kanji_off(kanji);
				outchar(c1);
			}
#else /* KANJI */
			outchar(c1);
#endif /* KANJI */
		}
#ifdef KANJI
		kanji = kanji_off(kanji);
#endif /* KANJI */
	}

	/* write new message */
	if (msg)
	{
		int	dist;
		u_char *cp;

		curs_columns();
		srow = Cursrow;
		scol = Curscol;

		slen = 0;
		cp = msg; 
		while(*cp && slen < MAXMSG && slen < Columns)
#ifdef KANJI
			if (IsKanji(*cp))
			{
				cp  += 2;
				slen+= 2;
			}
			else
#endif
			{
				++ cp;
				++ slen;
			}
		*cp = NUL;

		dist = p_fm > 0 ? p_fm : - p_fm;
		if (srow < Rows - dist - 1)
			srow += dist;
		else
			srow -= dist;

		if (srow < 0)    srow = 0;
		if (srow > Rows) srow = Rows - 1;

		screenp = LinePointers[srow];
		attribp = ScreenAttr(screenp);
		attr = rev ? AM_None : AM_Inv;

		/* find the location */
		{
			int fscol, bscol;
			u_char *cp, *sp;
			fscol = bscol = -1;
			cp = screenp;
											/* search forward */
			if (scol > slen)
				cp += scol - slen;

			for(; cp <= screenp + Columns - slen; cp++)
			{
				for(sp = cp; sp - cp != slen ; sp++)
					if (*sp != ' ')
					{
						cp = sp;
						break;
					}
				if (cp != sp)
				{
					fscol = cp - screenp;
					break;
				}
			}
											/* search backward */
		    for(cp = screenp + scol; cp >= screenp + slen; cp--)
			{
				for(sp = cp; cp - sp != slen; sp--)
					if (*sp != ' ')
					{
						cp = sp;
						break;
					}
				if (cp != sp)
				{
					bscol = sp - screenp + 1;
					break;
				}
			}

			if (fscol < 0)
				if (bscol < 0)		/* when no enough space is found */
					if (scol < Columns - slen - dist)
						scol += dist;
					else if (scol > slen + dist)
						scol -= slen + dist;
					else
					    scol = 0;
				else				/* when space is found in backword */
					scol = bscol;
			else
				if (bscol < 0) 		/* when space is found in foreword */
					scol = fscol;
				else
					scol = (scol - bscol -slen < fscol - scol) ? bscol : fscol;
		}

		screenp += scol;
		attribp += scol;
#ifdef KANJI
		kanji = kanji_off(kanji);

		if (IsKanji(*screenp) && !(*attribp & AM_K1))
		{
			screenp++;
			attribp++;
			scol++;
		}
#endif /* KANJI */

		osp = orgscr;
		oap = orgatr;
		for(mp = msg; *mp; mp++)
	 	{
			*osp ++ = *screenp;
			*oap ++ = *attribp;
			*screenp++ = *mp;
			*attribp++ = attr;
		}

		if (rev) outstr(T_TI);
		windgoto(srow, scol);

		for(mp = msg; *mp; mp++)
		{
			u_char c1, c2;
			c1 = *mp;
#ifdef KANJI
			if (IsKanji(c1))
			{
				c2 = * ++mp;
				kanji = kanji_on(kanji);
				kanjito(&c1, &c2, KANJI_DISP);
				outchar(c1);
				outchar(c2);
			}
			else
			{
				kanji = kanji_off(kanji);
				outchar(c1);
			}
#else /* KANJI */
			outchar(c1);
#endif /* KANJI */
		}
#ifdef KANJI
		kanji = kanji_off(kanji);
		if (IsKanji(*screenp) && !(*attribp & AM_K1))
		{
			*osp = *screenp;
			*oap = *attribp;
			*screenp = ' ';
			*attribp = AM_Inv;
			outchar(' ');
			slen++;
		}
#endif /* KANJI */
		if (rev) outstr(T_TP);
	}
}


